{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE OverloadedLabels #-}
{-# LANGUAGE TemplateHaskell #-}

import qualified Codec.Archive.Tar             as Tar

import           Control.Concurrent             ( forkIO )
import           Control.Concurrent.Async.Lifted
                                                ( async
                                                , cancel
                                                )
import           Control.Concurrent.Chan.Lifted ( Chan
                                                , newChan
                                                , readChan
                                                , writeChan
                                                )
import           Control.Concurrent.Lifted      ( threadDelay )
import           Control.Concurrent.MVar.Lifted ( MVar
                                                , newEmptyMVar
                                                , putMVar
                                                , readMVar
                                                , takeMVar
                                                )
import           Control.Lens
import           Control.Monad                  ( forever )
import           Control.Monad.Base             ( MonadBase )
import           Control.Monad.IO.Class         ( MonadIO
                                                , liftIO
                                                )
import           Control.Monad.Logger           ( MonadLogger
                                                , logError
                                                , logInfo
                                                , runChanLoggingT
                                                , runStdoutLoggingT
                                                , unChanLoggingT
                                                )
import           Control.Monad.Trans.Control    ( MonadBaseControl )
import           Data.List.NonEmpty             ( NonEmpty((:|)) )

import qualified Data.Conduit                  as C
import qualified Data.Conduit.Binary           as CB
import           Data.Default                   ( def )
import           Data.Generics.Labels           ( )
import           Data.Maybe                     ( fromMaybe )
import           Data.Monoid                    ( (<>) )
import           Data.Text                      ( Text )
import qualified Data.Text                     as T
import           Data.String                    ( fromString )

import           Network.Minio                  ( Bucket
                                                , Credentials(..)
                                                , ConnectInfo
                                                , GetObjectResponse
                                                , MinioErr
                                                , ObjectInfo(oiETag)
                                                , defaultGetObjectOptions
                                                , fromAWSEnv
                                                , fromMinioEnv
                                                , getObject
                                                , gorObjectInfo
                                                , gorObjectStream
                                                , runMinio
                                                , setCreds
                                                , statObject
                                                )
import           Network.Minio.S3API            ( ETag )
import           Network.Http.Csp               ( DirectiveList
                                                , Keyword(..)
                                                , SourceExpression(..)
                                                , Directive(..)
                                                , SchemePart(..)
                                                , HostPart(..)
                                                )
import           Network.Wai.Serve.Applications ( redirect )
import           Network.Wai.Serve.Main         ( serve' )
import           Network.Wai.Serve.Middleware   ( (<#>)
                                                , cspHeadersMiddleware
                                                , deindexifyMiddleware
                                                , domainMiddleware
                                                , forceSSLMiddleware
                                                , gzipMiddleware
                                                , loggerMiddleware
                                                , securityHeadersMiddleware
                                                , stsHeadersMiddleware
                                                )
import           Network.Wai.Serve.Types        ( Stage(..)
                                                , TLSConfiguration(..)
                                                , NotFoundHandlerConfig(..)
                                                , MiddlewareConfig(..)
                                                )

import           System.Directory               ( createDirectoryIfMissing
                                                , doesDirectoryExist
                                                , getCurrentDirectory
                                                , removeDirectoryRecursive
                                                , makeAbsolute
                                                )
import           System.Environment             ( lookupEnv )
import           System.FilePath.Posix          ( (</>) )

import           Espresso.Config                ( loadConfig
                                                , BundleConfig(..)
                                                , Config
                                                )

lookupEnvOrDefault :: (Read a, MonadIO m) => String -> a -> m a
lookupEnvOrDefault name def = do
  lookup <- liftIO $ lookupEnv name

  pure $ maybe def read lookup

-- | The entry point of the server application.
main :: IO ()
main = do
  config <- loadConfig

  loggerChan <- newChan
  restartChan <- newChan
  etagMVar <- newEmptyMVar

  -- Start a logger thread
  forkIO $ runStdoutLoggingT $ unChanLoggingT loggerChan

  runChanLoggingT loggerChan $ case (config ^. #bundle) of
    LocalBundle bundleConfig -> do
      path_ <- liftIO $ makeAbsolute (T.unpack $ bundleConfig ^. #path)

      $(logInfo)
        $ "Bundle: Using local path backend ("
        <> (T.pack path_)
        <> ")."

      liftIO $ runServer config path_
    S3Bundle bundleConfig -> do
      $(logInfo) "Bundle: Using object storage backend."

      cwd <- liftIO getCurrentDirectory

      let runDir = cwd </> "run/"
      let serveDir = runDir </> "serve/"
      let bundlePath = runDir </> (T.unpack bundleObjectName)

      liftIO $ createDirectoryIfMissing True runDir

      $(logInfo) ("Bucket: " <> bucket)

      minioResult <- downloadBundle cb3CI bucket bundlePath bundleObjectName

      case minioResult of
        Left e -> $(logError) $ "getObject failed." <> T.pack (show e)
        Right response -> do
          $(logInfo) "Sucessfully downloaded bundle."

          let bundleInfo = gorObjectInfo response
          $(logInfo) $ "ETag: " <> oiETag bundleInfo

          putMVar etagMVar (oiETag bundleInfo)

          async $ foreverServer restartChan config serveDir bundlePath

          forever $ updater cb3CI
                            bucket
                            etagMVar
                            restartChan
                            bundlePath
                            bundleObjectName
                            (bundleConfig ^. #pollSeconds)

     where
      bucket = bundleConfig ^. #bucket
      bundleObjectName = bundleConfig ^. #objectName
      endpointUrl = T.unpack $ bundleConfig ^. #endpoint
      cb3CI = setCreds
        (Credentials (bundleConfig ^. #accessKey) (bundleConfig ^. #secretKey))
        (fromString endpointUrl)

downloadBundle
  :: (MonadIO a, MonadLogger a)
  => ConnectInfo
  -> Bucket
  -> FilePath
  -> Text
  -> a (Either MinioErr GetObjectResponse)
downloadBundle connectInfo bucket bundlePath bundleObjectName = do
  $(logInfo) "Downloading latest bundle..."

  liftIO $ runMinio connectInfo $ do
    response <- getObject bucket bundleObjectName defaultGetObjectOptions

    C.connect (gorObjectStream response) $ CB.sinkFileCautious bundlePath

    pure response

updater
  :: (MonadIO m, MonadLogger m, MonadBase IO m)
  => ConnectInfo
  -> Bucket
  -> MVar ETag
  -> Chan Bool
  -> FilePath
  -> Text
  -> Int
  -> m ()
updater connectInfo bucket etagMVar restartChan bundlePath bundleObjectName pollSeconds
  = do
    threadDelay (1000000 * pollSeconds)

    $(logInfo) "Checking for bundle updates..."

    minioResult <- liftIO $ runMinio connectInfo $ statObject
      bucket
      bundleObjectName
      defaultGetObjectOptions

    case minioResult of
      Left e -> $(logError) $ "statObject failed." <> T.pack (show e)
      Right response -> do
        currentETag <- readMVar etagMVar
        let newETag = oiETag response

        if newETag == currentETag
          then $(logInfo) "No changes"
          else do
            $(logInfo) $ "Got new ETag: " <> oiETag response

            minioResult_ <- downloadBundle connectInfo
                                           bucket
                                           bundlePath
                                           bundleObjectName

            case minioResult_ of
              Left e -> $(logError) $ "getObject failed." <> T.pack (show e)
              Right response -> do
                $(logInfo) "Sucessfully downloaded new bundle."

                -- Update eTag.
                _ <- takeMVar etagMVar
                putMVar etagMVar newETag

                -- Schedule restart of web server.
                writeChan restartChan True

foreverServer
  :: (MonadIO m, MonadLogger m, MonadBaseControl IO m)
  => Chan Bool
  -> Config
  -> FilePath
  -> FilePath
  -> m ()
foreverServer restartChan config serveDir bundlePath = forever $ do
  serverId <- async $ prepareAndRunServer config serveDir bundlePath

  _ <- readChan restartChan

  $(logInfo) "Stopping server thread..."
  cancel serverId

prepareAndRunServer
  :: (MonadIO a, MonadLogger a) => Config -> FilePath -> FilePath -> a ()
prepareAndRunServer config serveDir bundlePath = do
  prepareServeDir serveDir

  $(logInfo) "Extracting bundle..."
  liftIO $ Tar.extract serveDir bundlePath

  liftIO $ runServer config serveDir

prepareServeDir :: (MonadIO a, MonadLogger a) => FilePath -> a ()
prepareServeDir serveDir = do
  serveDirExists <- liftIO $ doesDirectoryExist serveDir

  if serveDirExists
    then do
      $(logInfo) "Removing existing serve directory..."

      liftIO $ removeDirectoryRecursive serveDir
    else pure ()

  $(logInfo) "Recreating serve directory..."
  liftIO $ createDirectoryIfMissing True serveDir

runServer :: Config -> FilePath -> IO ()
runServer config serveDir = do
  let serverConfig = config ^. #server

  stage_ <- lookupEnvOrDefault "SITE_STAGE" (serverConfig ^. #stage)

  let serverEnvironmentConfig = case stage_ of
        Development -> serverConfig ^. #dev
        Staging -> serverConfig ^. #staging
        Production -> serverConfig ^. #prod

  serve' $ (serverEnvironmentConfig & #path .~ (Just serveDir))
